@echo off
py -m build
twine upload dist/* -u __token__ -p pypi-AgEIcHlwaS5vcmcCJGVlZWE0NjFhLTc2MTEtNGYxOS04NjAzLWE3YTA0MWU1NjNmNwACEVsxLFsibHVuYXJjb3JkIl1dAAIsWzIsWyI3ZjBiOTczNy02ODM2LTQxMzUtYjg5MC0wNjU1ZmM4NGQxYTIiXV0AAAYgw3djtPO8KlfZ-lipA6y9ZiRuI7GGrsdTNMy-mBBJIvA
pause >nul